import Foundation
import RxSwift
import Alamofire

/// Subclass of MoyaProvider that returns Observable instances when requests are made. Much better than using completion closures.
public class RxMoyaProvider<Target where Target: MoyaTarget>: MoyaProvider<Target> {
    /// Initializes a reactive provider.
    override public init(endpointClosure: EndpointClosure = MoyaProvider.DefaultEndpointMapping,
        requestClosure: RequestClosure = MoyaProvider.DefaultRequestMapping,
        stubClosure: StubClosure = MoyaProvider.NeverStub,
        manager: Manager = Alamofire.Manager.sharedInstance,
        plugins: [Plugin<Target>] = []) {
            super.init(endpointClosure: endpointClosure, requestClosure: requestClosure, stubClosure: stubClosure, manager: manager, plugins: plugins)
    }

    /// Designated request-making method.
    public func request(token: Target) -> Observable<MoyaResponse> {

        // Creates an observable that starts a request each time it's subscribed to.
        return create { [weak self] observer in
            let cancellableToken = self?.request(token) { data, statusCode, response, error in
                if let error = error {
                    observer.onError(error as NSError)
                } else {
                    if let data = data {
                        observer.onNext(MoyaResponse(statusCode: statusCode!, data: data, response: response))
                    }
                    observer.onCompleted()
                }
            }

            return AnonymousDisposable {
                cancellableToken?.cancel()
            }
        }
    }
}
